import re
from typing import Dict, List, Tuple, Optional
from dataclasses import dataclass


@dataclass
class PackageInfo:
    package_type: str
    package_name: str
    new_version: str
    old_version: str
    update_status: str
    repo: str

    def add_attributes(self, attrs):
        for attr_name, attr_value in attrs.items():
            setattr(self, attr_name, attr_value)


@dataclass
class UpdateSection:
    update_type: str
    update_status: bool
    update_details: Dict[str, PackageInfo]


@dataclass
class PretendError:
    error_type: str
    error_details: List[str]


@dataclass
class PretendSection:
    pretend_status: bool
    pretend_details: PretendError


@dataclass
class DiskUsageStats:
    mount_point: str
    total: str
    used: str
    free: str
    percent_used: str


@dataclass
class DiskUsage:
    before_update: List[DiskUsageStats]
    after_update: List[DiskUsageStats]


@dataclass
class LogInfo:
    pretend_emerge: Optional[PretendSection]
    update_system: Optional[UpdateSection]
    disk_usage: Optional[DiskUsage]


class Parser:
    def __init__(self, log_file: str) -> None:
        """
        Initialize the Parser with a log file.

        Parameters:
            log_file: The name of the log file.
        """
        self.log_file = log_file
        self.log_data = self.read_log()

    def read_log(self) -> List[str]:
        """
        Reads the log file and returns its content.

        Returns:
            List[str]: The content of the log file as a list of strings.
        """
        with open(self.log_file, "r") as log_file:
            log_data = log_file.readlines()
        return self.split_log_to_sections(log_data)

    def split_log_to_sections(self, log_data: List[str]) -> Dict:
        """
        Splits the log file into sections based on specified markers.

        Parameters:
            log_data (List[str]): Log content read with read_log method.

        Returns:
            Dict: A dictionary with section names as keys
                and content as values.
        """
        section_pattern = r"\{\{(.+?)\}\}"
        section_name = "beginning"
        log_by_sections: Dict[str, List[str]] = {section_name: []}

        for line in log_data:
            if " ::: " in line:
                line = line.split(" ::: ")[1].strip()
                match_section_pattern = re.search(section_pattern, line)
                if match_section_pattern:
                    section_name = "_".join(line.split()[1:-1]).lower()
                    log_by_sections[section_name] = []
                else:
                    log_by_sections[section_name].append(line)
            else:
                log_by_sections["final"] = line

        return log_by_sections

    def parse_emerge_pretend_section(
        self, section_content: List[str]
    ) -> PretendSection:
        """
        Function to parse the "emerge pretend" section of the log data.

        Parameters:
            section_content (List[str]): A list of strings that contains
                        the content of the "emerge pretend" section.

        Returns:
            Dict: A dictionary that contains the status of
                  the "emerge pretend" operation and the details.
        """
        if "emerge pretend was successful, updating..." in section_content:
            pretend_status = True
            pretend_details = None
        elif "There are no packages to update, skipping..." in section_content:
            pretend_status = True
            pretend_details = "No Updates"
        else:
            pretend_status = False
            pretend_details = self.parse_pretend_details(section_content)
        return PretendSection(pretend_status, pretend_details)

    def _parse_pretend_get_blocked_details(
        self, error_content: List[str]
    ) -> List[str]:
        """
        Parse details of blocked package error.

        Parameters:
            error_content (List[str]): Lines from log marked as ERROR by logger.

        Returns:
            List[str]: Parsed packages tht cause the error.
        """
        blocked_packages = []
        for line in error_content:
            match_package_pattern = re.search(r"^\((.+?)\)", line)
            if match_package_pattern:
                blocked_packages.append(match_package_pattern.group(1))

        return blocked_packages

    def _parse_pretend_get_error_type(
        self, error_content
    ) -> Tuple[str, str, List[str]]:
        """
        Parse error type and error details.

        Parameters:
            error_content (List[str]): Lines from log marked as ERROR by logger.

        Returns:
            Tuple[str, str, List[str]]: Return error type, error definition as strings
                and error details as a list of strings.
        """
        error_type = "undefined"
        error_definition = "undefined"
        error_details = "undefined"

        for line in error_content:
            if "Blocked Packages" in line:
                error_type = "Blocked Packages"
                error_details = self._parse_pretend_get_blocked_details(
                    error_content
                )

        for line in error_content:
            if line[0] == "*":
                error_definition += f"{line[2:]} "

        return (error_type, error_definition, error_details)

    def parse_pretend_details(self, section_content: List[str]) -> PretendError:
        """
        Parse information about the emerge pretend from logs.

        Parameters:
            section_content (List[str]): A list where each item is
                one line of logs from a section.

        Returns:
            PretendError: A datalass that contains details about failed pretend,
                for example a list of blocked packages.
        """
        error_index = section_content.index(
            "emerge pretend has failed, exiting"
        )
        error_content = [
            line for line in section_content[error_index + 1 :] if line
        ]

        error_type, _, error_details = self._parse_pretend_get_error_type(
            error_content
        )

        return PretendError(error_type, error_details)

    def parse_update_system_section(
        self, section_content: List[str]
    ) -> UpdateSection:
        """
        Function to parse the "update system" section of the log data.

        Parameters:
            section_content (List[str]): A list of strings that contains
                    the content of the "update system" section.

        Returns:
            UpdateSection: A dataclass that contains the status
                  of the system update and package details.
        """
        try:
            update_type = (
                "@world"
                if section_content[2].split()[1] == "@world"
                else "security"
            )
        except IndexError:
            update_type = "Undefined"

        if "update was successful" in section_content:
            update_status = True
            if "Nothing to merge; quitting" in section_content:
                update_details = {"updated_packages": [], "errors": []}
            else:
                package_list = self.parse_update_details(section_content)
                update_details = {"updated_packages": package_list}
        elif "There are no packages to update, skipping..." in section_content:
            update_status = True
            update_type = "security"
            update_details = {"updated_packages": [], "errors": []}
        else:
            update_status = False
            update_details = {"updated_packages": [], "errors": []}
        return UpdateSection(update_type, update_status, update_details)

    def parse_package_string(self, package_string: str) -> List[str]:
        """
        Parse package string into multiple parts.

        Parameters:
            package_string (str): String containing package information.

        Returns:
            List: List containing package information parts.
        """
        split_package_string = []
        temp = ""
        quotes_count = 0
        brackets_count = 0

        for char in package_string:
            temp += char
            if char == '"':
                quotes_count += 1
            elif char == "[":
                brackets_count += 1
            elif char == "]":
                brackets_count -= 1

            if char == " " and quotes_count % 2 == 0 and brackets_count == 0:
                split_package_string.append(temp.strip())
                temp = ""

        if temp:
            split_package_string.append(temp.strip())

        return split_package_string

    def parse_update_details(
        self, section_content: List[str]
    ) -> List[PackageInfo]:
        """
        Parse information about update from log file.

        Parameters:
            section_content (List[str]): A list where each item is
                one line of logs from a section.

        Returns:
            List[PackageInfo]: List of PackageInfo objects where each object
                contains useful information for the report.
        """
        ebuild_info_pattern = r"\[(.+?)\]"
        package_strings = [
            line
            for line in section_content
            if re.search(ebuild_info_pattern, line) and line != "[ ok ]"
        ]
        packages = []
        for package_string in package_strings:
            split_package_string = self.parse_package_string(package_string)
            update_status = split_package_string[0]

            if "ebuild" in update_status:
                package = self._parse_package_ebuild(split_package_string)
            elif "blocks" in update_status:
                package = self._parse_package_blocks(split_package_string)

            packages.append(package)

        return packages

    def _parse_package_ebuild(self, split_package_string: str) -> PackageInfo:
        """ """
        package_type = "ebuild"
        update_status = split_package_string[0]
        package_base_info = split_package_string[1]
        repo = package_base_info.split("::")[1]
        name_newversion = package_base_info.split("::")[0]

        package_name = ""
        for part in name_newversion.split("-"):
            if part.isnumeric() == True:
                pass
            elif "." in part or ":" in part:
                pass
            elif len(part) == 2 and part[0] == "r" and part[1].isnumeric():
                pass
            else:
                package_name += f"{part}-"

        new_version = name_newversion.replace(package_name, "")
        old_version = split_package_string[2].split("::")[0][1:]
        package_name = package_name[:-1]

        ebuild_info = PackageInfo(
            package_type,
            package_name,
            new_version,
            old_version,
            update_status,
            repo,
        )

        for var in split_package_string:
            if '="' in var:
                var = var.split("=")
                ebuild_info.add_attributes({var[0]: var[1][1:-1].split(" ")})

        return ebuild_info

    def _parse_package_blocks(self, split_package_string: str) -> PackageInfo:
        """ """
        package_type = "blocks"
        package_name = split_package_string[1][1:]
        new_version = None
        old_version = None
        update_status = split_package_string[0]
        repo = None

        blocks_info = PackageInfo(
            package_type,
            package_name,
            new_version,
            old_version,
            update_status,
            repo,
        )

        blocks_info.add_attributes(
            {"blocked_package": split_package_string[-1][:-1]}
        )
        return blocks_info

    def parse_disk_usage_info(
        self, section_content: List[str]
    ) -> List[DiskUsageStats]:
        """
        Get disk usage information.

        Parameters:
            section_content (List[str]): A list where each item is
                one line of logs from a section.

        Returns:
            List[DiskUsageStats]: A list of DiskUsageStats objects
                for each mount point containing statistics of disk usage.
        """
        mount_point_lines = [
            line for line in section_content if line[0:14] == "Disk usage for"
        ]
        mount_points = []

        for line in mount_point_lines:
            split_content = line.split(" ===> ")
            stats = split_content[1].split(", ")

            mount_point = split_content[0].replace("Disk usage for ", "")
            total = stats[0].split("=")[1]
            used = stats[1].split("=")[1]
            free = stats[2].split("=")[1]
            percent_used = stats[3].split("=")[1]

            disk_usage_stats = DiskUsageStats(
                mount_point, total, used, free, percent_used
            )
            mount_points.append(disk_usage_stats)

        return mount_points

    def extract_info_for_report(self) -> LogInfo:
        """
        Extract information about the update from the log file.

        Returns:
            LogInfo: Dataclass containing parsed data from all sections.
        """
        log_info = LogInfo
        disk_usage = DiskUsage
        for section in self.log_data.keys():
            section_content = self.log_data[section]
            if section == "pretend_emerge":
                log_info.pretend_emerge = self.parse_emerge_pretend_section(
                    section_content
                )
            elif section == "update_system":
                log_info.update_system = self.parse_update_system_section(
                    section_content
                )
            elif section == "calculate_disk_usage_1":
                disk_usage.before_update = self.parse_disk_usage_info(
                    section_content
                )
            elif section == "calculate_disk_usage_2":
                disk_usage.after_update = self.parse_disk_usage_info(
                    section_content
                )
        log_info.disk_usage = disk_usage
        return log_info



import os
import argparse
from typing import Tuple, List, Dict
from .shell_runner import ShellRunner
from .parser import Parser
from .reporter import Reporter
from .notifier import Notifier
from ._version import __version__

current_path = os.path.dirname(os.path.realpath(__file__))


def create_cli() -> argparse.Namespace:
    """
    Define CLI command flags using argparse.

    Returns:
        parser.parse_args: Argparse commands
    """
    formatter = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(
        description="Automate updates on Gentoo Linux.",
        formatter_class=formatter,
    )

    parser.add_argument(
        "-m",
        "--update-mode",
        default="security",
        choices=["security", "full"],
        help="Set the update mode.\n"
        "Options:\n"
        "* security: update only security patches (GLSA)\n"
        "* full: do a full @world update\n"
        "Default: security\n",
    )
    parser.add_argument(
        "-a",
        "--args",
        default="",
        help="Additional arguments to be passed when in 'full' update mode.\n"
        "Example:\n"
        "--args 'quiet-build=n color=y keep-going'",
    )
    parser.add_argument(
        "-c",
        "--config-update-mode",
        default="ignore",
        choices=["ignore", "merge"],
        help="Set the way new configurations are handled after an update.\n"
        "Options:\n"
        "* ignore: do not update configuration files at all.\n"
        "* merge: automatically merge changes in configuration files.\n"
        "Default: ignore\n",
    )
    parser.add_argument(
        "-d",
        "--daemon-restart",
        action="store_true",
        help="Set whether to restart services and daemons after an update.\n",
    )
    parser.add_argument(
        "-e",
        "--clean",
        action="store_true",
        help="Set whether to clean orphaned packaged after an update.\n",
    )
    parser.add_argument(
        "-l",
        "--read-logs",
        action="store_true",
        help="Set whether to read elogs after an update.\n",
    )
    parser.add_argument(
        "-n",
        "--read-news",
        action="store_true",
        help="Set whether to read news after an update.\n",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Do not show logs on the terminal screen.\n",
    )
    parser.add_argument(
        "-r",
        "--report",
        action="store_true",
        help="Show report or the last update log.\n",
    )
    parser.add_argument(
        "-s",
        "--send-report",
        default="none",
        choices=["irc", "email", "mobile", "none"],
        help="Send update report via IRC bot, email (SendGrid) or mobile app.\n"
        "Default: none\n",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print updater version.\n",
    )

    args = parser.parse_args()
    return args


def make_conf_reader() -> Dict:
    """
    Read /etc/portage/make.conf.

    Returns:
        Dict: Parameters in the key:value format.
        Example: {'COMMON_FLAGS': '"-O2 -pipe"'}
    """
    make_conf = {}
    with open("/etc/portage/make.conf", "r") as make_conf_raw:
        lines = make_conf_raw.read().splitlines()
    key, value = None, []
    for line in lines:
        line = line.strip()
        if not line or line.startswith("#"):
            continue
        if "=" in line:
            if key:
                make_conf[key] = " ".join(value)
            parts = line.split("=", 1)
            key = parts[0].strip()
            value = [parts[1].strip()]
        else:
            value.append(line)
    if key:
        make_conf[key] = " ".join(value)
    return make_conf


def initiate_log_directory(make_conf) -> Tuple[str, List[str]]:
    """
    Create log directory if it does not exist.
    If PORTAGE_LOGDIR is not set, use the default directory.

    Returns:
        str: Log directory path.
        List[str]: List of messages to be logged.
    """
    try:
        log_dir = make_conf["PORTAGE_LOGDIR"].replace('"', "")
    except KeyError:
        log_dir = ""

    log_dir_messages = []
    if log_dir == "":
        log_dir = "/var/log/portage/gentoo-update"
        log_dir_messages.append(
            f"PORTAGE_LOGDIR not set, using default: {log_dir}"
        )
    else:
        log_dir = log_dir.replace('"', "")
        log_dir = f"{log_dir}/gentoo-update"
        log_dir_messages.append(f"PORTAGE_LOGDIR set, using: {log_dir}")

    if not os.path.exists(log_dir):
        os.mkdir(log_dir)
        log_dir_messages.append(f"Created log directory: {log_dir}")

    return log_dir, log_dir_messages


def generate_last_report(log_dir: str) -> None:
    """
    Show report for the last log located in $PORTAGE_LOGDIR.

    Parameters:
        log_dir (str): Directory where gentoo_update stores logs.
    """
    files = os.listdir(log_dir)
    paths = [os.path.join(log_dir, basename) for basename in files]
    last_log = max(paths, key=os.path.getctime)
    update_info = Parser(last_log).extract_info_for_report()
    return Reporter(update_info)


def add_prefixes(args: str) -> str:
    """
    Function to add prefixes to a list of arguments passed to
    --update-mode full.

    Parameters:
    args_list (str): A string of space separated arguments without prefixes
        example: "quiet-build=n color=y keep-going"

    Returns:
    str: A new string of space separated arguments with added prefixes,
        example: "--quiet-build=n --color=y --keep-going"
    """
    args = args.split(" ")
    prefixed_args = []

    for arg in args:
        if len(arg) == 1:
            prefixed_args.append("-" + arg)
        else:
            prefixed_args.append("--" + arg)
    return " ".join(prefixed_args)


def main() -> None:
    args = create_cli()

    make_conf = make_conf_reader()
    log_dir, log_dir_messages = initiate_log_directory(make_conf)

    if args.report:
        generate_last_report(log_dir).print_report()
    elif args.send_report in ["irc", "email", "mobile"]:
        report = generate_last_report(log_dir).create_report()
        short = False if args.send_report != "irc" else True
        Notifier(notification_type=args.send_report, report=report, short=short)
    else:
        runner = ShellRunner(
            "y" if args.quiet else "n", log_dir, log_dir_messages
        )
        runner.run_shell_script(
            args.update_mode,
            add_prefixes(args.args) if args.args else "NOARGS",
            args.config_update_mode,
            "y" if args.daemon_restart else "n",
            "y" if args.clean else "n",
            "y" if args.read_logs else "n",
            "y" if args.read_news else "n",
        )


if __name__ == "__main__":
    main()



from typing import List
from .parser import (
    UpdateSection,
    PretendError,
    PretendSection,
    DiskUsage,
    LogInfo,
)


class Reporter:
    def __init__(self, update_info: LogInfo) -> None:
        self.info = update_info

    def _create_failed_pretend_report(
        self, pretend_info: PretendSection
    ) -> List[str]:
        """
        Create a report when emerge pretend fails.

        Parameters:
            pretend_info (PretendSection): emerge pretend information.

        Returns:
            List: A list of strings that comprise the failed pretend report.
        """
        report = [
            "==========> Gentoo Update Report <==========",
            "emerge pretend status: FAIL",
        ]
        pretend_details = pretend_info.pretend_details
        if pretend_details.error_type == "undefined":
            report.append("Could not identify error, please check the logs")
        elif pretend_details.error_type == "Blocked Packages":
            report.append(f"\nError Type: {pretend_details.error_type}")
            report.extend(self._report_blocked_packages(pretend_details))
        report.append("")
        return report

    def _report_blocked_packages(
        self, pretend_details: PretendError
    ) -> List[str]:
        """
        Report on Blocked Packages error during emerge pretend.

        Parameters:
            pretend_details (PretendError): In this case it's blocked packages.

        Returns:
            List[str]: Section of the report about blocked packages.
        """
        blocked_packages_report = ["List of Blocked Packages:"]
        for package in pretend_details.error_details:
            blocked_packages_report.append(f"-----> {package}")

        return blocked_packages_report

    def _create_failed_report(
        self, update_info: UpdateSection, disk_usage_info: DiskUsage
    ) -> List[str]:
        """
        Create a report when update fails.

        Parameters:
            update_info (LogInfo.UpdateSection): Update information.
            disk_usage_inf (LogInfo.DiskUsage): Disk usage information.

        Returns:
            List: A list of strings that comprise the failed update report.
        """
        # do failed report processing
        return ["Your update is a failure"]

    def _create_successful_report(
        self, update_info: UpdateSection, disk_usage_info: DiskUsage
    ) -> List[str]:
        """
        Create a report when update succeeds.

        Parameters:
            update_info (LogInfo.UpdateSection): Update information.
            disk_usage_inf (LogInfo.DiskUsage): Disk usage information.

        Returns:
            List: A list of strings that comprise the successful update report.
        """
        report = [
            "==========> Gentoo Update Report <==========",
            "update status: SUCCESS",
        ]
        updated_packages = update_info.update_details["updated_packages"]
        other_package_types = False
        if updated_packages:
            report.append(r"processed packages:")
            for package in updated_packages:
                if package.package_type == "ebuild":
                    package_name = package.package_name
                    new_version = package.new_version
                    old_version = package.old_version
                    report.append(
                        f"--- {package_name} {old_version}->{new_version}"
                    )
                else:
                    other_package_types = True

            if other_package_types:
                report.append("")
                report.append("Non-ebuild packages")
                for package in updated_packages:
                    if package.package_type == "blocks":
                        package_name = package.package_name
                        blocked_package = package.blocked_package
                        report.append(
                            f"--- {package_name} blocked {blocked_package}"
                        )
            report.append("")
            report.append("Disk Usage Stats:")

            for before, after in zip(
                disk_usage_info.before_update, disk_usage_info.after_update
            ):
                disk_usage_stats = (
                    f"Mount Point {before.mount_point}\n"
                    f"Free Space {before.free} => {after.free}\n"
                    f"Used Space {before.used} => {after.used}\n"
                    f"Used pc(%) {before.percent_used} => {after.percent_used}\n"
                )
                report.append(disk_usage_stats)

        return report

    def create_report(self) -> List[str]:
        """
        Create a report.

        Returns:
            List: A list of strings that comprise the update report.
        """
        info = self.info
        disk_usage_info = info.disk_usage
        pretend_success = info.pretend_emerge.pretend_status

        if pretend_success:
            update_info = info.update_system
            update_success = info.update_system.update_status
        else:
            pretend_info = info.pretend_emerge
            update_success = False

        if update_success:
            report = self._create_successful_report(
                update_info, disk_usage_info
            )
        elif pretend_success and not update_success:
            report = self._create_failed_report(update_info, disk_usage_info)
        else:
            report = self._create_failed_pretend_report(pretend_info)
        return report

    def print_report(self) -> None:
        """
        Print the report line by line to console.
        """
        report = self.create_report()
        for line in report:
            print(line)



import unittest
from unittest.mock import patch, Mock
from gentoo_update import create_cli, add_prefixes


class TestGentooUpdate(unittest.TestCase):
    def test_add_prefixes(self):
        input_args = "a bc"
        expected_output = "-a --bc"
        self.assertEqual(add_prefixes(input_args), expected_output)

    @patch("argparse.ArgumentParser.parse_args")
    def test_create_cli(self, mock_parse_args):
        mock_args = Mock()
        mock_parse_args.return_value = mock_args
        args = create_cli()
        self.assertEqual(args, mock_args)


if __name__ == "__main__":
    unittest.main()
